//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "UaMain.h"
#include "UaConf.h"
#include "ControlQueue.h"
#include "Command.h"
#include "CallbackQueue.h"
#include "Callback.h"
#include "Settings.h"
#include <re.h>
#include "ua/baresip.h"
#include "core.h"
#include "Log.h"
#include "Utils.h"
#include <assert.h>

#pragma package(smart_init)

//---------------------------------------------------------------------------
class Worker : public TThread
{
private:
protected:
	void __fastcall Execute();
public:
	__fastcall Worker(bool CreateSuspended);
};


#define DEBUG_MODULE "UaMain"
#define DEBUG_LEVEL 6
#include <re_dbg.h>

static struct {
	const char *peeruri;  /**< Optional Peer URI               */
	uint32_t n_uas;       /**< Number of User Agents           */
	bool prefer_ipv6;     /**< Prefer IPv6 transport flag      */
	bool run_daemon;      /**< Run process as daemon flag      */
	bool terminating;     /**< Application is terminating flag */
} app;

static const char software[] = "tSIP 0.1";

static bool appRestart = false;
static bool appQuit = false;

static void ua_exit_handler(int err)
{
	(void)err;
	re_cancel();
}

static void ua_event_handler(enum ua_event ev, const char *prm, void *arg)
{
	Callback::ua_state_e state;
	Callback::reg_state_e reg_state;
	/** \todo acc_id */
	switch (ev)
	{
	case UA_EVENT_CALL_INCOMING:
		state = Callback::CALL_STATE_INCOMING;
		UA_CB->ChangeCallState(state, prm);
		break;
	case UA_EVENT_CALL_RINGING:
		state = Callback::CALL_STATE_RINGING;
		UA_CB->ChangeCallState(state, prm);
		break;
	case UA_EVENT_CALL_PROGRESS:
		state = Callback::CALL_STATE_PROGRESS;
		UA_CB->ChangeCallState(state, prm);
		break;
	case UA_EVENT_CALL_ESTABLISHED:
		state = Callback::CALL_STATE_ESTABLISHED;
		UA_CB->ChangeCallState(state, prm);
		break;
	case UA_EVENT_CALL_CLOSED:
		state = Callback::CALL_STATE_CLOSED;
		UA_CB->ChangeCallState(state, prm);
		break;
	case UA_EVENT_REGISTERING:
		reg_state = Callback::REG_STATE_REGISTERING;
		UA_CB->ChangeRegState(reg_state, 0);
		break;
	case UA_EVENT_REGISTER_OK:
		reg_state = Callback::REG_STATE_REGISTER_OK;
		UA_CB->ChangeRegState(reg_state, 0);
		break;
	case UA_EVENT_REGISTER_FAIL:
		reg_state = Callback::REG_STATE_REGISTER_FAIL;
		UA_CB->ChangeRegState(reg_state, 0);
		break;
	case UA_EVENT_UNREGISTERING:
		reg_state = Callback::REG_STATE_UNREGISTERING;
		UA_CB->ChangeRegState(reg_state, 0);
		break;
	case UA_EVENT_UNREGISTER_OK:
		reg_state = Callback::REG_STATE_UNREGISTER_OK;
		UA_CB->ChangeRegState(reg_state, 0);
		break;
	case UA_EVENT_UNREGISTER_FAIL:
		reg_state = Callback::REG_STATE_UNREGISTER_FAIL;
		UA_CB->ChangeRegState(reg_state, 0);
		break;
	default:
		assert(!"Unhandled UA event");
	}

}

static int ua_add(const struct pl *addr)
{
	struct ua *ua;
	char buf[512];
	int err;

	pl_strcpy(addr, buf, sizeof(buf));

	err = ua_alloc(&ua, buf, ua_event_handler, NULL, NULL);
	if (err)
		return err;

	/* skip status-line in daemon mode */
//	if (app.run_daemon)
//		ua_set_statmode(ua, STATMODE_OFF);
	ua_set_statmode(ua, STATMODE_OFF);

	return err;
}

extern "C" void on_log(int level, const char *p)
{
	LOG("%s", p);
}

static int app_init(void)
{
	char uuid[37] = "";
	uint32_t n;
	int err;

    memset(&app, 0, sizeof(app));

	print_handler_set(on_log);

	/* Initialise System library */
	err = libre_init();
	if (err)
		return err;

	dbg_init(DBG_INFO, DBG_NONE);

	/* Initialise dynamic modules */
	mod_init();

	static AnsiString asConfPath = ExtractFileDir(Application->ExeName).c_str();
	conf_path_set(asConfPath.c_str());

    config.sound_backend = (conf::e_sound_backend)appSettings.uaConf.soundBackend;

	config.codecs.g711 = appSettings.uaConf.codecs.g711;
	config.codecs.g726_32 = appSettings.uaConf.codecs.g726_32;
	config.codecs.gsm = appSettings.uaConf.codecs.gsm;
	config.codecs.speex = appSettings.uaConf.codecs.speex;

	/* Parse configuration */
	err = configure();
	if (err) {
		DEBUG_WARNING("configure: %s\n", strerror(err));
		return err;
	}

	n = list_count(aucodec_list());
	LOG("Populated %u audio codec%s\n", n, 1==n?"":"s");
	if (0 == n) {
		DEBUG_WARNING("No audio-codec modules loaded!\n");
	}
	n = list_count(aufilt_list());
	LOG("Populated %u audio filter%s\n", n, 1==n?"":"s");

	/* Initialise Network */
	err = net_init(false);	// do not prefer IP v6
	if (err) {
		DEBUG_WARNING("network init failed (%s)\n", strerror(err));
		return err;
	}

	/* Initialise User Agents */
	//TODO: version info
	AnsiString uaName;
	uaName.sprintf("tSIP %s", GetFileVer(Application->ExeName).c_str());
	err = ua_init(uaName.c_str(), appSettings.uaConf.logMessages,
		true, true, true, ua_exit_handler);
	if (err)
		return err;

	return 0;
}


static void net_change_handler(void *arg)
{
	(void)arg;

	re_printf("IP-address changed: %j\n", net_laddr());

	ua_reset_transp(true, true);
}


static int app_start(void)
{
	uint32_t n;
	int err;

	app.n_uas = 0;
	for (int i=0; i<appSettings.uaConf.accounts.size(); i++)
	{
		AnsiString addr;
		UaConf::Account acc = appSettings.uaConf.accounts[i];
		addr.sprintf("<sip:%s:%s@%s;transport=%s>;regint=%d", //;sipnat=outbound",
			acc.user.c_str(), acc.pwd.c_str(), acc.reg_server.c_str(),
			acc.getTransportStr(),
			acc.reg_expires
			);
		if (acc.auth_user != "")
		{
			addr.cat_printf(";auth_user=%s", acc.auth_user.c_str());
		}

		if (acc.stun_server != "")
		{
			addr.cat_printf(";medianat=stun;stunserver=stun:%s",
				acc.stun_server.c_str());
		}

		pl pl_addr;
		pl_set_str(&pl_addr, addr.c_str());
		if (ua_add(&pl_addr) == 0)
		{
			app.n_uas++;
		}
	}
	if (!app.n_uas)
	{
		DEBUG_WARNING("No SIP accounts found - check your config\n");
		return ENOENT;
	}

	/* Start all User-Agents */
	err = ua_start_all();
	if (err)
	{
		return err;
	}

	/* Populate contacts */
	contact_init();
	//err = conf_contacts_get(contact_add, &n);

	net_change(60, net_change_handler, NULL);

	return 0;
}


static void app_close(void)
{
	ua_close();
	contact_close();
	net_close();
	play_close();
	audio_loop_test(true);
#ifdef USE_VIDEO
	video_loop_test(true);
#endif
	mod_close();
	libre_close();

	/* Check for memory leaks */
    LOG("Checking for memory leaks...");
	tmr_debug();
	mem_debug();
}


static void quit(int sig)
{
	if (!app.terminating) {
		ua_stop_all(false);
		app.terminating = true;
		return;
	}

	LOG("forced exit - signal %d\n", sig);

	ua_stop_all(true);
}


extern "C" void signal_handler(int sig)
{
	LOG("terminated by signal %d\n", sig);
	quit(sig);
}


extern "C" int stderr_handler(const char *p, size_t size, void *arg)
{
	(void)arg;

	if (1 != fwrite(p, size, 1, stderr))
		return ENOMEM;

	return 0;
}

extern "C" void control_handler(void)
{
	if (app.terminating)
	{
		return;
	}
	else if (appQuit || appRestart)
	{
		quit(0);
	}
	
	int err;
	Command cmd;
	if (UA->GetCommand(cmd))
	{
    	return;
	}
	switch (cmd.type)
	{
	case Command::CALL:
		err = ua_connect(ua_cur(), cmd.target.c_str(), NULL, NULL, VIDMODE_OFF);
		if (err)
		{
			DEBUG_WARNING("connect failed: %s\n", strerror(err));
		}
		break;
	case Command::ANSWER:
		ua_answer(ua_cur());
		break;
	case Command::SEND_DIGIT:
		ua_play_digit(ua_cur(), cmd.key);
		break;
	case Command::HANGUP:
		ua_hangup(ua_cur());
		break;
	case Command::SET_MSG_LOGGING:
		ua_log_messages(cmd.bEnabled);
		break;
	default:
		assert(!"Unhandled command type");
		break;
	}
}

static bool app_terminated = false;

__fastcall Worker::Worker(bool CreateSuspended)
	: TThread(CreateSuspended)
{
}
//---------------------------------------------------------------------------
void __fastcall Worker::Execute()
{
	int err;

	do
	{
		appRestart = false;
		err = app_init();
		if (err) {
			DEBUG_WARNING("app_init failed (%s)\n", strerror(err));
			goto out;
		}

		err = app_start();
		if (err) {
			DEBUG_WARNING("app_start failed (%s)\n", strerror(err));
			UA_CB->ChangeAppState(Callback::APP_START_FAILED);
			goto out;
		}

		if (app.run_daemon) {
			DEBUG_NOTICE("Starting in daemon mode\n");
			err = sys_daemon();
			if (err) {
				DEBUG_WARNING("daemon: %s\n", strerror(err));
				goto out;
			}
		}

		/* Automatically call peer uri if set */
		if (app.peeruri) {
			err = ua_connect(ua_cur(), app.peeruri, NULL,
					 NULL, VIDMODE_OFF);
			if (err) {
				DEBUG_WARNING("connect failed: %s\n", strerror(err));
			}
		}

		/* Main loop */
		while (!app.terminating) {

			err = re_main(signal_handler, control_handler);

			if (!app.terminating)
				ua_stack_resume(software, appSettings.uaConf.logMessages, true, true, true);
		}

	 out:
		app_close();

		if (err) {
			DEBUG_WARNING("main exit with error (%s)\n", strerror(err));
		} else {
			LOG("main exit w/o error, waiting for restart or termination\n");
		}
		while (!appRestart && !appQuit && !Terminated)
		{
			Sleep(10);
		}
	} while (appRestart);
	//return err;
	app_terminated = true;
}
//---------------------------------------------------------------------------

int Ua::Start(void)
{
	thread = new Worker(false);
	thread->FreeOnTerminate = true;
	return 0;
}

void Ua::Restart(void)
{
	appRestart = true;
}

void Ua::Quit(void)
{
	appQuit = true;
	while (!app_terminated)
	{
    	Sleep(10);
	}
}



